@@ -1,50 +1,61 @@
- # Copyright (c) 2000 Gregory Trubetskoy.  All rights reserved.
+ # ====================================================================
+ # The Apache Software License, Version 1.1
+ #
+ # Copyright (c) 2000-2002 The Apache Software Foundation.  All rights
+ # reserved.
  #
  # Redistribution and use in source and binary forms, with or without
  # modification, are permitted provided that the following conditions
  # are met:
  #
  # 1. Redistributions of source code must retain the above copyright
- #    notice, this list of conditions and the following disclaimer. 
+ #    notice, this list of conditions and the following disclaimer.
  #
  # 2. Redistributions in binary form must reproduce the above copyright
  #    notice, this list of conditions and the following disclaimer in
  #    the documentation and/or other materials provided with the
  #    distribution.
  #
- # 3. The end-user documentation included with the redistribution, if
- #    any, must include the following acknowledgment: "This product 
- #    includes software developed by Gregory Trubetskoy."
- #    Alternately, this acknowledgment may appear in the software itself, 
+ # 3. The end-user documentation included with the redistribution,
+ #    if any, must include the following acknowledgment:
+ #       "This product includes software developed by the
+ #        Apache Software Foundation (http://www.apache.org/)."
+ #    Alternately, this acknowledgment may appear in the software itself,
  #    if and wherever such third-party acknowledgments normally appear.
  #
- # 4. The names "mod_python", "modpython" or "Gregory Trubetskoy" must not 
- #    be used to endorse or promote products derived from this software 
- #    without prior written permission. For written permission, please 
- #    contact grisha@modpython.org.
+ # 4. The names "Apache" and "Apache Software Foundation" must
+ #    not be used to endorse or promote products derived from this
+ #    software without prior written permission. For written
+ #    permission, please contact apache@apache.org.
  #
- # 5. Products derived from this software may not be called "mod_python"
- #    or "modpython", nor may "mod_python" or "modpython" appear in their 
- #    names without prior written permission of Gregory Trubetskoy.
+ # 5. Products derived from this software may not be called "Apache",
+ #    "mod_python", or "modpython", nor may these terms appear in their
+ #    name, without prior written permission of the Apache Software
+ #    Foundation.
  #
- # THIS SOFTWARE IS PROVIDED BY GREGORY TRUBETSKOY ``AS IS'' AND ANY
- # EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- # IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- # PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL GREGORY TRUBETSKOY OR
- # HIS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- # SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- # NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- # HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
- # STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
- # OF THE POSSIBILITY OF SUCH DAMAGE.
+ # THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
+ # WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ # OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ # DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
+ # ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ # SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ # LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
+ # USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+ # ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ # OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
+ # OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ # SUCH DAMAGE.
  # ====================================================================
  #
- # $Id: apache.py,v 1.34.4.2 2002/04/10 21:18:22 gtrubetskoy Exp $
+ # This software consists of voluntary contributions made by many
+ # individuals on behalf of the Apache Software Foundation.  For more
+ # information on the Apache Software Foundation, please see
+ # <http://www.apache.org/>.
+ #
+ # Originally developed by Gregory Trubetskoy.
+ #
 
 import sys
-import string
 import traceback
 import time
 import os
@@ -58,44 +69,6 @@ import _apache
 # variable stores the last PythonPath in raw (unevaled) form.
 _path = None
 
-# this is used in Request.__init__
-def _cleanup_request(_req):
-    _req._Request = None
-
-class Request:
-    """ This is a normal Python Class that can be subclassed.
-        However, most of its functionality comes from a built-in
-        object of type mp_request provided by mod_python at
-        initialization and stored in self._req.
-    """
-
-    def __init__(self, _req):
-        # look at __setattr__ if changing this line!
-        self._req = _req
-        
-        # this will decrement the reference to the _req
-        # object at cleanup time. If we don't do this, we
-        # get a cirular reference and _req never gets destroyed.
-        _req.register_cleanup(_cleanup_request, _req)
-
-    def __getattr__(self, attr):
-        try:
-            return getattr(self._req, attr)
-        except AttributeError:
-            raise AttributeError, attr
-
-    def __setattr__(self, attr, val):
-        try:
-            if attr != "_req":
-                setattr(self._req, attr, val)
-            else:
-                self.__dict__["_req"] = val
-        except AttributeError:
-            self.__dict__[attr] = val
-
-    def __nonzero__(self):
-        return 1
-
 class CallBack:
     """
     A generic callback object.
@@ -112,51 +85,223 @@ class CallBack:
 
         def pop(self):
 
-            handlers = string.split(self.req.hstack)
+            handlers = self.req.hstack.split()
+
             if not handlers:
                 return None
             else:
-                self.req.hstack = string.join(handlers[1:], " ")
+                self.req.hstack = " ".join(handlers[1:])
                 return handlers[0]
 
-    def Dispatch(self, _req, htype):
+    def ConnectionDispatch(self, conn):
+
+        # config
+        config = conn.base_server.get_config()
+        debug = config.has_key("PythonDebug")
+
+        try:
+
+            handler = conn.hlist.handler
+            
+            # split module::handler
+            l = handler.split('::', 1)
+            module_name = l[0]
+            if len(l) == 1:
+                # no oject, provide default
+                object_str = "connectionhandler"
+            else:
+                object_str = l[1]
+
+            # add the directory to pythonpath if
+            # not there yet, or pythonpath specified
+            
+            if config.has_key("PythonPath"):
+                # we want to do as little evaling as possible,
+                # so we remember the path in un-evaled form and
+                # compare it
+                global _path
+                pathstring = config["PythonPath"]
+                if pathstring != _path:
+                    _path = pathstring
+                    newpath = eval(pathstring)
+                    if sys.path != newpath:
+                        sys.path[:] = newpath
+            else:
+                if filter.dir not in sys.path:
+                    sys.path[:0] = [filter.dir]
+
+            # import module
+            module = import_module(module_name, config)
+
+            # find the object
+            object = resolve_object(module, object_str,
+                                    arg=conn, silent=0)
+
+            if object:
+
+                # call the object
+                if config.has_key("PythonEnablePdb"):
+                    result = pdb.runcall(object, conn)
+                else:
+                    result = object(conn)
+
+                assert (type(result) == type(int())), \
+                       "ConnectionHandler '%s' returned invalid return code." % handler
+
+        except PROG_TRACEBACK, traceblock:
+            # Program run-time error
+            try:
+                (etype, value, traceback) = traceblock
+                result = self.ReportError(etype, value, traceback, srv=conn.base_server,
+                                          phase="ConnectionHandler",
+                                          hname=handler, debug=debug)
+            finally:
+                traceback = None
+
+        except:
+            # Any other rerror (usually parsing)
+            try:
+                exc_type, exc_value, exc_traceback = sys.exc_info()
+                filter.disable()
+                result = self.ReportError(exc_type, exc_value, exc_traceback, srv=conn.base_server,
+                                          phase=filter.name, hname=handler, debug=debug)
+            finally:
+                exc_traceback = None
+
+	return result
+
+    def FilterDispatch(self, filter):
+
+        req = filter.req
+
+        # config
+        config = req.get_config()
+        debug = config.has_key("PythonDebug")
+
+        try:
+
+            # split module::handler
+            l = filter.handler.split('::', 1)
+            module_name = l[0]
+            if len(l) == 1:
+                # no oject, provide default
+                if filter.is_input:
+                    object_str = "inputfilter"
+                else:
+                    object_str = "outputfilter"
+            else:
+                object_str = l[1]
+
+            # add the directory to pythonpath if
+            # not there yet, or pythonpath specified
+            
+            if config.has_key("PythonPath"):
+                # we want to do as little evaling as possible,
+                # so we remember the path in un-evaled form and
+                # compare it
+                global _path
+                pathstring = config["PythonPath"]
+                if pathstring != _path:
+                    _path = pathstring
+                    newpath = eval(pathstring)
+                    if sys.path != newpath:
+                        sys.path[:] = newpath
+            else:
+                if filter.dir not in sys.path:
+                    sys.path[:0] = [filter.dir]
+
+            # import module
+            module = import_module(module_name, config)
+
+            # find the object
+            object = resolve_object(module, object_str, arg=filter, silent=0)
+
+            if object:
+
+                # call the object
+                if config.has_key("PythonEnablePdb"):
+                    pdb.runcall(object, filter)
+                else:
+                    object(filter)
+
+                # always flush the filter. without a FLUSH or EOS bucket,
+                # the content is never written to the network.
+                # XXX an alternative is to tell the user to flush() always
+                filter.flush()
+
+        except SERVER_RETURN, value:
+            # SERVER_RETURN indicates a non-local abort from below
+            # with value as (result, status) or (result, None) or result
+            try:
+                if len(value.args) == 2:
+                    (result, status) = value.args
+                    if status:
+                        req.status = status
+                else:
+                    result = value.args[0]
+                    
+                if type(result) != type(7):
+                    s = "Value raised with SERVER_RETURN is invalid. It is a "
+                    s = s + "%s, but it must be a tuple or an int." % type(result)
+                    _apache.log_error(s, APLOG_NOERRNO|APLOG_ERR, req.server)
+
+                    return
+
+            except:
+                pass
+
+        except PROG_TRACEBACK, traceblock:
+            # Program run-time error
+            try:
+                (etype, value, traceback) = traceblock
+                filter.disable()
+                result = self.ReportError(etype, value, traceback, req=req, filter=filter,
+                                          phase="Filter: " + filter.name,
+                                          hname=filter.handler, debug=debug)
+            finally:
+                traceback = None
+
+        except:
+            # Any other rerror (usually parsing)
+            try:
+                exc_type, exc_value, exc_traceback = sys.exc_info()
+                filter.disable()
+                result = self.ReportError(exc_type, exc_value, exc_traceback, req=req, filter=filter,
+                                          phase=filter.name, hname=filter.handler,
+                                          debug=debug)
+            finally:
+                exc_traceback = None
+
+	return OK
+
+    def HandlerDispatch(self, req):
         """
         This is the handler dispatcher.
         """
 
         # be cautious
-        result, handler = HTTP_INTERNAL_SERVER_ERROR, ""
-
-        # is there a Request object for this request?
-        if not _req._Request:
-            _req._Request = Request(_req)
-            
-        req = _req._Request
+        result = HTTP_INTERNAL_SERVER_ERROR
 
         # config
-        config = _req.get_config()
+        config = req.get_config()
         debug = config.has_key("PythonDebug")
 
         try:
+            hlist = req.hlist
 
-            # cycle through the handlers
-            dirs = _req.get_all_dirs()
-
-            hstack = self.HStack(_req)
-
-            handler = hstack.pop()
-            while handler:
+            while hlist.handler:
 
                 # split module::handler
-                l = string.split(handler, '::', 1)
+                l = hlist.handler.split('::', 1)
+
                 module_name = l[0]
                 if len(l) == 1:
                     # no oject, provide default
-                    object_str = string.lower(htype[len("python"):])
+                    object_str = req.phase[len("python"):].lower()
                 else:
                     object_str = l[1]
-                
-                # add the direcotry to pythonpath if
+
+                # add the directory to pythonpath if
                 # not there yet, or pythonpath specified
                 if config.has_key("PythonPath"):
                     # we want to do as little evaling as possible,
@@ -170,19 +315,16 @@ class CallBack:
                         if sys.path != newpath:
                             sys.path[:] = newpath
                 else:
-                    if config.has_key("PythonHandlerModule"):
-                        dir = _req.get_all_dirs()["PythonHandlerModule"]
-                    else:
-                        dir = _req.get_all_dirs()[htype]
+                    dir = hlist.directory
                     if dir not in sys.path:
                         sys.path[:0] = [dir]
 
                 # import module
-                module = import_module(module_name, _req)
+                module = import_module(module_name, config)
 
                 # find the object
-                silent = config.has_key("PythonHandlerModule")
-                object = resolve_object(req, module, object_str, silent)
+                object = resolve_object(module, object_str,
+                                        arg=req, silent=hlist.silent)
 
                 if object:
 
@@ -192,30 +334,33 @@ class CallBack:
                     else:
                         result = object(req)
 
+                    assert (type(result) == type(int())), \
+                           "Handler '%s' returned invalid return code." % hlist.handler
+
                     # stop cycling through handlers
                     if result != OK:
                         break
                     
-                elif silent:
+                elif hlist.silent:
                     result = DECLINED
-                        
-                handler = hstack.pop()
+
+                hlist.next()
 
         except SERVER_RETURN, value:
-            # SERVER_RETURN indicates a non-local abort from below
+            # SERVER_RETURN indicates an abort from below
             # with value as (result, status) or (result, None) or result
             try:
                 if len(value.args) == 2:
                     (result, status) = value.args
                     if status:
-                        _req.status = status
+                        req.status = status
                 else:
                     result = value.args[0]
                     
                 if type(result) != type(7):
                     s = "Value raised with SERVER_RETURN is invalid. It is a "
                     s = s + "%s, but it must be a tuple or an int." % type(result)
-                    _apache.log_error(s, APLOG_NOERRNO|APLOG_ERR, _req.server)
+                    _apache.log_error(s, APLOG_NOERRNO|APLOG_ERR, req.server)
                     return HTTP_INTERNAL_SERVER_ERROR
 
             except:
@@ -225,8 +370,8 @@ class CallBack:
             # Program run-time error
             try:
                 (etype, value, traceback) = traceblock
-                result = self.ReportError(req, etype, value, traceback,
-                                          htype=htype, hname=handler,
+                result = self.ReportError(etype, value, traceback, req=req,
+                                          phase=req.phase, hname=hlist.handler,
                                           debug=debug)
             finally:
                 traceback = None
@@ -235,81 +380,100 @@ class CallBack:
             # Any other rerror (usually parsing)
             try:
                 exc_type, exc_value, exc_traceback = sys.exc_info()
-                result = self.ReportError(req, exc_type, exc_value, exc_traceback,
-                                          htype=htype, hname=handler, debug=debug)
+                result = self.ReportError(exc_type, exc_value, exc_traceback, req=req,
+                                          phase=req.phase, hname=hlist.handler, debug=debug)
             finally:
                 exc_traceback = None
 
 	return result
 
 
-    def ReportError(self, req, etype, evalue, etb, htype="N/A", hname="N/A", debug=0):
+    def ReportError(self, etype, evalue, etb, req=None, filter=None, srv=None,
+                    phase="N/A", hname="N/A", debug=0):
 	""" 
 	This function is only used when debugging is on.
 	It sends the output similar to what you'd see
 	when using Python interactively to the browser
 	"""
 
-        try:
+        try:       # try/finally
+            try:        # try/except
 
-            if str(etype) == "exceptions.IOError" \
-               and str(evalue)[:5] == "Write":
-                # if this is an IOError while writing to client,
-                # it is probably better not to try to write to the cleint
-                # even if debug is on.
-                debug = 0
-            
-            # write to log
-            for e in traceback.format_exception(etype, evalue, etb):
-                s = "%s %s: %s" % (htype, hname, e[:-1])
-                _apache.log_error(s, APLOG_NOERRNO|APLOG_ERR, req.server)
+                if str(etype) == "exceptions.IOError" \
+                   and str(evalue)[:5] == "Write":
+                    # if this is an IOError while writing to client,
+                    # it is probably better not to try to write to the cleint
+                    # even if debug is on.
+                    debug = 0
 
-            if not debug:
-                return HTTP_INTERNAL_SERVER_ERROR
-            else:
-                # write to client
-
-                req.content_type = 'text/plain'
-                req.send_http_header()
-
-                s = '\nMod_python error: "%s %s"\n\n' % (htype, hname)
+                # write to log
                 for e in traceback.format_exception(etype, evalue, etb):
-                    s = s + e + '\n'
+                    s = "%s %s: %s" % (phase, hname, e[:-1])
+                    if req:
+                        req.log_error(s, APLOG_NOERRNO|APLOG_ERR)
+                    else:
+                        _apache.log_error(s, APLOG_NOERRNO|APLOG_ERR, srv)
 
-                req.write(s)
+                if not debug or not req:
+                    return HTTP_INTERNAL_SERVER_ERROR
+                else:
+                    # write to client
+                    req.content_type = 'text/plain'
 
-                return DONE
+                    s = '\nMod_python error: "%s %s"\n\n' % (phase, hname)
+                    for e in traceback.format_exception(etype, evalue, etb):
+                        s = s + e + '\n'
+                        
+                    if filter:
+                        filter.write(s)
+                        filter.flush()
+                    else:
+                        req.write(s)
+
+                    return DONE
+            except:
+                # last try
+                traceback.print_exc()
 
         finally:
             # erase the traceback
             etb = None
             # we do not return anything
 
-def import_module(module_name, req=None, path=None):
+def import_module(module_name, config=None, path=None):
     """
     Get the module to handle the request. If
     autoreload is on, then the module will be reloaded
     if it has changed since the last import.
     """
- 
+
     # Get options
     debug, autoreload = 0, 1
-    if req:
-        config = req.get_config()
+    if config:
         debug = config.has_key("PythonDebug")
         if config.has_key("PythonAutoReload"):
             autoreload = int(config["PythonAutoReload"])
- 
+
     # (Re)import
     if sys.modules.has_key(module_name):
-        
+
         # The module has been imported already
         module = sys.modules[module_name]
 
         # but is it in the path?
         file = module.__dict__.get("__file__")
-        if not file or (path and not file in path):
-                raise SERVER_RETURN, HTTP_NOT_FOUND
+
+        # the "and not" part of this condition is to prevent execution
+        # of arbitrary already imported modules, such as os. the
+        # not-so-obvious filter(lambda...) call means:
+        # return entries from path (which is a list) which fully match at
+        # beginning the dirname of file. E.g. if file is /a/b/c/d.py, a path
+        # of /a/b/c will pass, because the file is below the /a/b/c path, but
+        # path of /a/b/c/g will not pass.
+
+        if (not file or path and not
+            filter(lambda a: os.path.dirname(file).find(a) == 0, path)):
+            raise SERVER_RETURN, HTTP_NOT_FOUND
 
         if autoreload:
             oldmtime = module.__dict__.get("__mtime__", 0)
@@ -319,19 +483,22 @@ def import_module(module_name, req=None, path=None):
 
     else:
         mtime, oldmtime = 0, -1
- 
+
     if mtime > oldmtime:
 
         # Import the module
         if debug:
-            s = 'mod_python: (Re)importing %s from %s' % (module_name, path)
+            if path:
+                s = "mod_python: (Re)importing module '%s' with path set to '%s'" % (module_name, path)
+            else:
+                s = "mod_python: (Re)importing module '%s'" % module_name
             _apache.log_error(s, APLOG_NOERRNO|APLOG_NOTICE)
 
-        parts = string.split(module_name, '.')
+        parts = module_name.split('.')
         for i in range(len(parts)):
             f, p, d = imp.find_module(parts[i], path)
             try:
-                mname = string.join(parts[:i+1], ".")
+                mname = ".".join(parts[:i+1])
                 module = imp.load_module(mname, f, p, d)
             finally:
                 if f: f.close()
@@ -342,7 +509,7 @@ def import_module(module_name, req=None, path=None):
             mtime = module_mtime(module)
 
         module.__mtime__ = mtime
- 
+
     return module
 
 def module_mtime(module):
@@ -367,7 +534,7 @@ def module_mtime(module):
 
     return mtime
 
-def resolve_object(req, module, object_str, silent=0):
+def resolve_object(module, object_str, arg=None, silent=0):
     """
     This function traverses the objects separated by .
     (period) to find the last one we're looking for:
@@ -375,11 +542,14 @@ def resolve_object(req, module, object_str, silent=0):
        From left to right, find objects, if it is
        an unbound method of a class, instantiate the
        class passing the request as single argument
+
+    'arg' is sometimes req, sometimes filter,
+    sometimes connection
     """
 
     obj = module
 
-    for obj_str in  string.split(object_str, '.'):
+    for obj_str in  object_str.split('.'):
 
         parent = obj
 
@@ -398,7 +568,7 @@ def resolve_object(req, module, object_str, silent=0):
         if hasattr(obj, "im_self") and not obj.im_self:
             # this is an unbound method, its class
             # needs to be instantiated
-            instance = parent(req)
+            instance = parent(arg)
             obj = getattr(instance, obj_str)
 
     return obj
@@ -411,9 +581,7 @@ def build_cgi_env(req):
     """
 
     req.add_common_vars()
-    env = {}
-    for k in req.subprocess_env.keys():
-        env[k] = req.subprocess_env[k]
+    env = req.subprocess_env.copy()
         
     if len(req.path_info) > 0:
         env["SCRIPT_NAME"] = req.uri[:-len(req.path_info)]
@@ -437,7 +605,7 @@ class NullIO:
     def readlines(self): return []
     def write(self, s): pass
     def writelines(self, list):
-        self.write(string.joinfields(list, ''))
+        self.write("".join(list))
     def isatty(self): return 0
     def flush(self): pass
     def close(self): pass
@@ -476,7 +644,7 @@ class CGIStdin(NullIO):
             return s
 
     def readlines(self):
-        s = string.split(self.buf + self.read(), '\n')
+        s = (self.buf + self.read()).split('\n')
         return map(lambda s: s + '\n', s)
 
     def readline(self, n = -1):
@@ -488,7 +656,7 @@ class CGIStdin(NullIO):
         self.buf = self.buf + self.req.read(self.BLOCK)
 
         # look for \n in the buffer
-        i = string.find(self.buf, '\n')
+        i = self.buf.find('\n')
         while i == -1: # if \n not found - read more
             if (n != -1) and (len(self.buf) >= n): # we're past n
                 i = n - 1
@@ -498,7 +666,7 @@ class CGIStdin(NullIO):
             if len(self.buf) == x: # nothing read, eof
                 i = x - 1
                 break 
-            i = string.find(self.buf, '\n', x)
+            i = self.buf.find('\n', x)
         
         # carve out the piece, then shorten the buffer
         result = self.buf[:i+1]
@@ -530,10 +698,10 @@ class CGIStdout(NullIO):
             headers_over = 0
 
             # first try RFC-compliant CRLF
-            ss = string.split(self.headers, '\r\n\r\n', 1)
+            ss = self.headers.split('\r\n\r\n', 1)
             if len(ss) < 2:
                 # second try with \n\n
-                ss = string.split(self.headers, '\n\n', 1)
+                ss = self.headers.split('\n\n', 1)
                 if len(ss) >= 2:
                     headers_over = 1
             else:
@@ -541,21 +709,20 @@ class CGIStdout(NullIO):
                     
             if headers_over:
                 # headers done, process them
-                string.replace(ss[0], '\r\n', '\n')
-                lines = string.split(ss[0], '\n')
+                ss[0] = ss[0].replace('\r\n', '\n')
+                lines = ss[0].split('\n')
                 for line in lines:
-                    h, v = string.split(line, ":", 1)
-                    v = string.strip(v)
-                    if string.lower(h) == "status":
-                        status = int(string.split(v)[0])
+                    h, v = line.split(":", 1)
+                    v = v.strip()
+                    if h.lower() == "status":
+                        status = int(v.split()[0])
                         self.req.status = status
-                    elif string.lower(h) == "content-type":
+                    elif h.lower() == "content-type":
                         self.req.content_type = v
                         self.req.headers_out[h] = v
                     else:
                         self.req.headers_out.add(h, v)
-                self.req.send_http_header()
-                self.headers_sent = 1
+
                 # write the body if any at this point
                 self.req.write(ss[1])
         else:
@@ -574,32 +741,29 @@ def setup_cgi(req):
     """
 
     # save env
-    env = os.environ.copy()
+    save_env = os.environ.copy()
     
     si = sys.stdin
     so = sys.stdout
 
-    env = build_cgi_env(req)
+    os.environ.update(build_cgi_env(req))
  
-    for k in env.keys():
-        os.environ[k] = env[k]
-
     sys.stdout = CGIStdout(req)
     sys.stdin = CGIStdin(req)
 
     sys.argv = [] # keeps cgi.py happy
 
-    return env, si, so
+    return save_env, si, so
         
-def restore_nocgi(env, si, so):
+def restore_nocgi(sav_env, si, so):
     """ see setup_cgi() """
 
     osenv = os.environ
 
     # restore env
-    for k in osenv.keys():
+    for k in osenv:
         del osenv[k]
-    for k in env.keys():
+    for k in sav_env:
         osenv[k] = env[k]
 
     sys.stdout = si
@@ -613,10 +777,11 @@ def init():
     return CallBack()
 
 ## Some functions made public
-make_table = _apache.make_table
+make_table = _apache.table
 log_error = _apache.log_error
-parse_qs = _apache.parse_qs
-parse_qsl = _apache.parse_qsl
+table = _apache.table
+config_tree = _apache.config_tree
+server_root = _apache.server_root
 
 ## Some constants
 
@@ -704,16 +869,76 @@ REMOTE_NAME = 1
 REMOTE_NOLOOKUP = 2
 REMOTE_DOUBLE_REV = 3
 
+# legacy/mod_python things
 REQ_ABORTED = HTTP_INTERNAL_SERVER_ERROR
 REQ_EXIT = "REQ_EXIT"         
 SERVER_RETURN = _apache.SERVER_RETURN
 PROG_TRACEBACK = "PROG_TRACEBACK"
 
+# the req.finfo tuple
+FINFO_MODE = 0
+FINFO_INO = 1
+FINFO_DEV = 2
+FINFO_NLINK = 4
+FINFO_UID = 5
+FINFO_GID = 6
+FINFO_SIZE = 7
+FINFO_ATIME = 8
+FINFO_MTIME = 9
+FINFO_CTIME = 10
+FINFO_FNAME = 11
+FINFO_NAME = 12
+#FINFO_FILEHAND = 14
 
+# the req.parsed_uri
+URI_SCHEME = 0
+URI_HOSTINFO = 1
+URI_USER = 2
+URI_PASSWORD = 3
+URI_HOSTNAME = 4
+URI_PORT = 5
+URI_PATH = 6
+URI_QUERY = 7
+URI_FRAGMENT = 8
 
+# for req.proxyreq
+PROXYREQ_NONE = 0       # No proxy
+PROXYREQ_PROXY = 1	# Standard proxy
+PROXYREQ_REVERSE = 2	# Reverse proxy
 
+# methods for req.allow_method()
+M_GET = 0               # RFC 2616: HTTP
+M_PUT = 1
+M_POST = 2
+M_DELETE = 3
+M_CONNECT = 4
+M_OPTIONS = 5
+M_TRACE = 6             # RFC 2616: HTTP
+M_PATCH = 7 
+M_PROPFIND = 8          # RFC 2518: WebDAV
+M_PROPPATCH = 9         
+M_MKCOL = 10
+M_COPY = 11
+M_MOVE = 12
+M_LOCK = 13
+M_UNLOCK = 14           # RFC2518: WebDAV
+M_VERSION_CONTROL = 15  # RFC3253: WebDAV Versioning
+M_CHECKOUT = 16
+M_UNCHECKOUT = 17
+M_CHECKIN = 18
+M_UPDATE = 19
+M_LABEL = 20
+M_REPORT = 21
+M_MKWORKSPACE = 22
+M_MKACTIVITY = 23
+M_BASELINE_CONTROL = 24
+M_MERGE = 25
+M_INVALID = 26           # RFC3253: WebDAV Versioning
 
-
+# for req.used_path_info
+AP_REQ_ACCEPT_PATH_INFO = 0  # Accept request given path_info
+AP_REQ_REJECT_PATH_INFO = 1  # Send 404 error if path_info was given
+AP_REQ_DEFAULT_PATH_INFO = 2 # Module's choice for handling path_info
 
 
 
